import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.util.*;

public final class Main {

  public static final int SQRT_SAMPLES = 1;
  public static final int INVERSE_SAMPLES = 1 / (SQRT_SAMPLES * SQRT_SAMPLES);
  public static final int WIDTH = 640;
  public static final int HEIGHT = 512;
  public static final String IMAGE_TYPE = "jpg";
  public static final String OUTPUT_FILE = "output." + IMAGE_TYPE;
  public static final int NOISE_SIZE = 1 << 8;
  public static final int NOISE_MASK = NOISE_SIZE - 1;
  public static final long NOISE_SEED = 0xDEADBEEFCAFEBABEL;
  public static final int FBM_ITERATIONS = 16;
  public static final int SKY_PALETTE_SIZE = 256;
  public static final double SKY_PALETTE_SCALE = 20;
  public static final double SKY_PALETTE_OFFSET = -12;

  public static final int SKY_BLUE = 0x3D66C2;
  public static final int CLOUD_WHITE = 0xFDFEF8;

  public static final int RED = 0;
  public static final int GREEN = 1;
  public static final int BLUE = 2;

  public static final int FBM_GAIN = 0;
  public static final int FBM_LACUNARITY = 1;
  public static final int FBM_SCALE = 2;

  public static final double GAMMA = 2.2;
  public static final double INVERSE_GAMMA = 1.0 / GAMMA;

  public static final long SECOND_MILLIS = 1000L;
  public static final long MINUTE_MILLIS = 60 * SECOND_MILLIS;
  public static final long HOUR_MILLIS = 60 * MINUTE_MILLIS;

  private double[][][] pixels = new double[HEIGHT][WIDTH][3];

  private double[][][] noise = new double[NOISE_SIZE][NOISE_SIZE][NOISE_SIZE];
  private double[] skyBlue = new double[3];
  private double[] cloudWhite = new double[3];
  private double[][] skyPalette = new double[SKY_PALETTE_SIZE][3];
  private double[] fbmSkyMaterial = { 0.4, 2.5, 0.005 };

  private void launch() throws Throwable {

    initializeNoise();
    initializeSkyPalette();

    for(int a = 0, z = 0; a < 8; a++) {
      for(int b = 0; b < 8; b++, z += 10) {
        for(int y = 0; y < HEIGHT / 8; y++) {
          for(int x = 0; x < WIDTH / 8; x++) {
            int index = toPaletteIndex(
                fractionalBrownianMotion(x * 8, y * 8, z, fbmSkyMaterial));
            for(int i = 0; i < 3; i++) {
              pixels[a * HEIGHT / 8 + y][b * WIDTH / 8 + x][i]
                  = skyPalette[index][i];
            }
          }
        }
      }
    }


    saveImage();

  }

  private void initializeSkyPalette() {
    convertHexColorToDoubles(cloudWhite, CLOUD_WHITE);
    convertHexColorToDoubles(skyBlue, SKY_BLUE);
    for(int i = 0; i < SKY_PALETTE_SIZE; i++) {
      double value = 0.5 + Math.atan(SKY_PALETTE_OFFSET + SKY_PALETTE_SCALE * i
          / (double)(SKY_PALETTE_SIZE - 1)) / (Math.PI / 2) / 2;
      if (value < 0) {
        value = 0;
      } else if (value > 1) {
        value = 1;
      }
      for(int j = 0; j < 3; j++) {
        skyPalette[i][j] = cloudWhite[j] + (skyBlue[j] - cloudWhite[j]) * value;
      }
    }
  }

  private void convertHexColorToDoubles(double[] output, int input) {
    for(int i = 0; i < 3; i++) {
      int hex = input & 0xFF;
      input >>= 8;
      output[2 - i] = Math.pow(hex / 255.0, GAMMA);
    }
  }

  private int toPaletteIndex(double intensity) {
    int index = (int)Math.round(255 * intensity);
    if (index < 0) {
      index = 0;
    } else if (index > 255) {
      index = 255;
    }
    return index;
  }

  private double fractionalBrownianMotion(
      double x, double y, double z, double[] fbmMaterial) {

    double gain = 1.0;
    double lacunarity = fbmMaterial[FBM_SCALE];
    double sum = 0.0;

    for(int i = 0; i < FBM_ITERATIONS;
        i++, gain *= fbmMaterial[FBM_GAIN],
        lacunarity *= fbmMaterial[FBM_LACUNARITY]) {
      sum += gain * noiseFunction(
          x * lacunarity, y * lacunarity, z * lacunarity);
    }

    return sum;
  }

  private double noiseFunction(double x, double y, double z) {

    int floorX = (int)Math.floor(x);
    int floorY = (int)Math.floor(y);
    int floorZ = (int)Math.floor(z);

    double percentX = x - floorX;
    double percentY = y - floorY;
    double percentZ = z - floorZ;

    int x1 = floorX & NOISE_MASK;
    int y1 = floorY & NOISE_MASK;
    int z1 = floorZ & NOISE_MASK;
    int x2 = (x1 + 1) & NOISE_MASK;
    int y2 = (y1 + 1) & NOISE_MASK;
    int z2 = (z1 + 1) & NOISE_MASK;

    double n111 = noise[z1][y1][x1];
    double n211 = noise[z1][y1][x2];
    double n121 = noise[z1][y2][x1];
    double n221 = noise[z1][y2][x2];

    double n112 = noise[z2][y1][x1];
    double n212 = noise[z2][y1][x2];
    double n122 = noise[z2][y2][x1];
    double n222 = noise[z2][y2][x2];

    double n11 = n111 + (n112 - n111) * percentZ;
    double n21 = n211 + (n212 - n211) * percentZ;
    double n12 = n121 + (n122 - n121) * percentZ;
    double n22 = n221 + (n222 - n221) * percentZ;

    double a = n11 + (n12 - n11) * percentY;
    double b = n21 + (n22 - n21) * percentY;

    return a + (b - a) * percentX;
  }

  private void initializeNoise() {
    Random random = new Random(NOISE_SEED);
    for(int i = 0; i < NOISE_SIZE; i++) {
      for(int j = 0; j < NOISE_SIZE; j++) {
        for(int k = 0; k < NOISE_SIZE; k++) {
          noise[i][j][k] = random.nextDouble();
        }
      }
    }
  }

  private void saveImage() throws Throwable {
    int[] data = new int[WIDTH * HEIGHT];

    for(int y = 0, k = 0; y < HEIGHT; y++) {
      for(int x = 0; x < WIDTH; x++, k++) {
        int value = 0;
        for(int i = 0; i < 3; i++) {
          int intensity = (int)Math.round(255
              * Math.pow(pixels[y][x][i] * INVERSE_SAMPLES, INVERSE_GAMMA));
          if (intensity < 0) {
            intensity = 0;
          } else if (intensity > 255) {
            intensity = 255;
          }
          value <<= 8;
          value |= intensity;
        }
        data[k] = value;
      }
    }

    BufferedImage image = new BufferedImage(
        WIDTH, HEIGHT, BufferedImage.TYPE_INT_RGB);
    image.setRGB(0, 0, WIDTH, HEIGHT, data, 0, WIDTH);
    ImageIO.write(image, IMAGE_TYPE, new File(OUTPUT_FILE));
  }

  public static void main(String... args) throws Throwable {

    long startTime = System.currentTimeMillis();

    Main main = new Main();
    main.launch();

    long interval = System.currentTimeMillis() - startTime;
    long hours = interval / HOUR_MILLIS;
    interval %= HOUR_MILLIS;
    long minutes = interval / MINUTE_MILLIS;
    interval %= MINUTE_MILLIS;
    long seconds = interval / SECOND_MILLIS;
    interval %= SECOND_MILLIS;
    System.out.format("%d hour%s, %d minute%s, %d second%s, %d millisecond%s%n",
        hours, hours == 1 ? "" : "s",
        minutes, minutes == 1 ? "" : "s",
        seconds, seconds == 1 ? "" : "s",
        interval, interval == 1 ? "" : "s");
  }

}
